open Big_int

let report = ref false;;
let ivy = ref false
let sexpr = ref false
let latex = ref false
let fancy = ref false
let html = ref false
let prover9style = ref false
let tptpstyle = ref false
    
let ostr x =
  match x with
  | None -> ""
  | Some(y) -> y
   
let esc s =
  let b = Buffer.create 10 in
  for i = 0 to (String.length s) - 1 do
    if s.[i] = '\\' then
      Buffer.add_string b "\\\\"
    else
      Buffer.add_char b s.[i]
  done;
  Buffer.contents b

let int_of_hexchar c =
  match c with
  | '0' -> 0
  | '1' -> 1
  | '2' -> 2
  | '3' -> 3
  | '4' -> 4
  | '5' -> 5
  | '6' -> 6
  | '7' -> 7
  | '8' -> 8
  | '9' -> 9
  | 'a' -> 10
  | 'b' -> 11
  | 'c' -> 12
  | 'd' -> 13
  | 'e' -> 14
  | 'f' -> 15
  | _ -> raise (Failure "non hex char")

let hex_of_big_int x len =
  let b = Buffer.create (len*2) in
  for i = len-1 downto 0 do
    Buffer.add_string b (Printf.sprintf "%02x" (int_of_big_int (mod_big_int (shift_right_towards_zero_big_int x (i*8)) (big_int_of_int 256))))
  done;
  Buffer.contents b

let big_int_of_hex x =
  let r = ref zero_big_int in
  for i = 0 to (String.length x) - 1 do
    r := add_int_big_int (int_of_hexchar x.[i]) (mult_int_big_int 16 !r) 
  done;
  !r
    
let split_string splitchar x =
  let rl = ref [] in
  let b = Buffer.create 64 in
  for i = 0 to (String.length x) - 1 do
    let c = x.[i] in
    if c = splitchar then
      begin
	rl := Buffer.contents b::!rl;
	Buffer.clear b;
      end
    else
      Buffer.add_char b c
  done;
  let l = Buffer.contents b in
  rl := l::!rl;
  List.rev !rl

(* base58 representation *)
let _base58strs = ["1";"2";"3";"4";"5";"6";"7";"8";"9";"A";"B";"C";"D";"E";"F";"G";"H";"J";"K";"L";"M";"N";"P";"Q";"R";"S";"T";"U";"V";"W";"X";"Y";"Z";"a";"b";"c";"d";"e";"f";"g";"h";"i";"j";"k";"m";"n";"o";"p";"q";"r";"s";"t";"u";"v";"w";"x";"y";"z"]

(* c : big_int *)
(* return base58 string representation of c *)
let rec base58_rec c r =
  if gt_big_int c zero_big_int then
    let (q,m) = quomod_big_int c (big_int_of_string "58") in
    base58_rec q ((List.nth _base58strs (int_of_big_int m)) ^ r)
  else
    r

let base58 c = base58_rec c ""

let base58char_int c =
   match c with
   | '1' -> 0
   | '2' -> 1
   | '3' -> 2
   | '4' -> 3
   | '5' -> 4
   | '6' -> 5
   | '7' -> 6
   | '8' -> 7
   | '9' -> 8
   | 'A' -> 9
   | 'B' -> 10
   | 'C' -> 11
   | 'D' -> 12
   | 'E' -> 13
   | 'F' -> 14
   | 'G' -> 15
   | 'H' -> 16
   | 'J' -> 17
   | 'K' -> 18
   | 'L' -> 19
   | 'M' -> 20
   | 'N' -> 21
   | 'P' -> 22
   | 'Q' -> 23
   | 'R' -> 24
   | 'S' -> 25
   | 'T' -> 26
   | 'U' -> 27
   | 'V' -> 28
   | 'W' -> 29
   | 'X' -> 30
   | 'Y' -> 31
   | 'Z' -> 32
   | 'a' -> 33
   | 'b' -> 34
   | 'c' -> 35
   | 'd' -> 36
   | 'e' -> 37
   | 'f' -> 38
   | 'g' -> 39
   | 'h' -> 40
   | 'i' -> 41
   | 'j' -> 42
   | 'k' -> 43
   | 'm' -> 44
   | 'n' -> 45
   | 'o' -> 46
   | 'p' -> 47
   | 'q' -> 48
   | 'r' -> 49
   | 's' -> 50
   | 't' -> 51
   | 'u' -> 52
   | 'v' -> 53
   | 'w' -> 54
   | 'x' -> 55
   | 'y' -> 56
   | 'z' -> 57
   | _ -> raise (Failure "bad base58 char")

(* s : base58 string *)
(* return int representation of s *)
let rec frombase58_rec s i sl r =
  if i < sl then
    frombase58_rec s (i + 1) sl (add_int_big_int (base58char_int (String.get s i)) (mult_int_big_int 58 r))
  else
    r

let frombase58 s = frombase58_rec s 0 (String.length s) zero_big_int
